package auth

import (
	"context"
	"log"
	"net/http"
)

type contextKey string

// User is an application user.
type User struct {
	ID      int    `db:"rowid"`
	Name    string `db:"name"`
	IsAdmin bool   `db:"is_admin"`
}

type Provider interface {
	// Valid checks whether the given user and pass combination represents a valid user. If so, it returns
	// a user value. If the combination is not a valid user, or another error occurs, Valid returns an error.
	Valid(ctx context.Context, user, pass string) (*User, error)
}

// Require wraps hdlr so that it requires authentication to use. Requests handled by hdlr will have the user
// name attached to their context. Use the User function to retrieve it.
func Require(hdlr http.HandlerFunc, provider Provider) http.HandlerFunc {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		userName, pass, ok := r.BasicAuth()
		if !ok {
			w.Header().Add("WWW-Authenticate", `Basic realm="In Vino Veritas"`)
			w.WriteHeader(http.StatusUnauthorized)

			return
		}

		user, err := provider.Valid(r.Context(), userName, pass)
		if err != nil {
			log.Printf("can't authenticate %s %s from %s: %s", r.Method, r.URL, r.RemoteAddr, err)

			w.Header().Add("WWW-Authenticate", `Basic realm="In Vino Veritas"`)
			w.WriteHeader(http.StatusUnauthorized)

			return
		}

		if user == nil {
			w.Header().Add("WWW-Authenticate", `Basic realm="In Vino Veritas"`)
			w.WriteHeader(http.StatusUnauthorized)

			return
		}

		key := contextKey("user")
		ctx := context.WithValue(r.Context(), key, *user)

		hdlr(w, r.WithContext(ctx))
	})
}

// Get returns the user from r's context. If there is no user in r's context, returns an empty user.
func Get(r *http.Request) User {
	val := r.Context().Value(contextKey("user"))
	if val == nil {
		return User{}
	}

	return val.(User)
}
