package main

import (
	"bytes"
	"context"
	"errors"
	"fmt"
	"image"
	"image/png"
	"io"
	"log"

	// Imported for side effects to register format handlers
	_ "image/jpeg"
	_ "image/png"

	"github.com/Masterminds/squirrel"
	"github.com/jmoiron/sqlx"
	"golang.org/x/image/draw"
)

type storeOperation int

const (
	Add storeOperation = iota
	Update
)

type Comment struct {
	Content string `db:"content"`
}

type Vino struct {
	ID         int         `db:"rowid"`
	Name       string      `db:"name"`
	Rating     int         `db:"rating"`
	Picture    image.Image `db:"-"`
	HasPicture bool        `db:"has_picture"` // Set to true if there's picture data for this vino
	Comments   []Comment   `db:"-"`
}

func DeleteVino(ctx context.Context, db *sqlx.DB, id int) error {
	query, args, err := squirrel.Delete("wines").
		Where(squirrel.Eq{"rowid": id}).
		ToSql()
	if err != nil {
		return err
	}

	_, err = db.ExecContext(ctx, query, args...)
	if err != nil {
		return err
	}

	return nil
}

func LoadVino(ctx context.Context, db *sqlx.DB, id int) (Vino, error) {
	cols := []string{
		"rowid", "name", "rating",
		"CASE WHEN picture IS NOT NULL THEN true ELSE false END AS has_picture",
	}

	query, args, err := squirrel.Select(cols...).
		From("wines").
		Where(squirrel.Eq{"rowid": id}).
		ToSql()
	if err != nil {
		return Vino{}, err
	}

	tx, err := db.Beginx()
	if err != nil {
		return Vino{}, err
	}
	defer tx.Rollback() // The tx is readonly anyway, no need to commit anything

	var v Vino
	err = tx.GetContext(ctx, &v, query, args...)
	if err != nil {
		return v, err
	}

	query, args, err = squirrel.Select("content").
		From("comments").
		Where(squirrel.Eq{"wine": id}).
		OrderBy("rowid ASC").
		ToSql()
	if err != nil {
		return v, err
	}

	err = tx.SelectContext(ctx, &v.Comments, query, args...)
	if err != nil {
		return v, err
	}

	return v, nil
}

func LoadPictureData(ctx context.Context, db *sqlx.DB, id int) ([]byte, error) {
	query := `SELECT picture FROM wines WHERE rowid = ?`

	var img []byte
	err := db.GetContext(ctx, &img, query, id)
	if err != nil {
		return nil, err
	}

	return img, nil
}

// AddPicture loads picture data (PNG or JPEG) from fh and sets v's picture to it.
// If something goes wrong during loading, or the image is neither PNG nor JPEG, an error
// is returned. If contentType is not the empty string, it is validated to be either
// image/png or image/jpeg.
func (v *Vino) AddPicture(fh io.Reader, contentType string) error {
	switch contentType {
	case "", "image/jpeg", "image/png":
	default:
		return fmt.Errorf("unexpected content type for image: %q", contentType)
	}

	img, _, err := image.Decode(fh)
	if err != nil {
		return err
	}

	v.Picture = img

	return nil
}

func (v Vino) String() string {
	return fmt.Sprintf("{Name: %q, Rating: %d}", v.Name, v.Rating)
}

func (v *Vino) Store(ctx context.Context, db *sqlx.DB, op storeOperation) (err error) {
	if op == Add && v.ID != 0 {
		// If we already have an ID, the op needs to be an Update
		return errors.New("add with already-set ID")
	}

	values := map[string]interface{}{
		"name":   v.Name,
		"rating": v.Rating,
	}

	log.Println("image data:", values)

	if v.Picture != nil {
		// Scale image down and encode as PNG
		// Get aspect ratio of incoming picture
		bounds := v.Picture.Bounds()
		aspect := float64(bounds.Max.X) / float64(bounds.Max.Y)

		const destHeight = 800

		rect := image.Rect(0, 0, int(destHeight*aspect), destHeight)

		log.Println("bounds:", bounds, "-> aspect:", aspect, "new size:", rect)

		scaled := image.NewRGBA(rect)

		draw.ApproxBiLinear.Scale(scaled, rect, v.Picture, v.Picture.Bounds(), draw.Over, nil)

		// Encode scaled image as PNG
		var img bytes.Buffer

		err := png.Encode(&img, scaled)
		if err != nil {
			return err
		}

		values["picture"] = img.Bytes()
	}

	tx, err := db.Begin()
	if err != nil {
		return err
	}

	defer func() {
		if err != nil {
			log.Println("rolling back transaction:", err)
			tx.Rollback()
			return
		}

		err = tx.Commit()
		if err != nil {
			log.Println("can't commit transaction:", err)
		}
	}()

	if op == Update {
		// Looks like we need to do an update instead
		_, err = squirrel.Update("wines").
			Where("rowid = ?", v.ID).
			SetMap(values).
			RunWith(tx).
			ExecContext(ctx)
		return err
	}

	res, err := squirrel.Insert("wines").
		SetMap(values).
		RunWith(tx).
		ExecContext(ctx)
	if err != nil {
		return err
	}

	id, err := res.LastInsertId()
	if err != nil {
		return err
	}

	// If it's an add (i.e. initial creation of an entry) fill in the ID
	log.Println("insert result:", res, "id:", id)

	v.ID = int(id)

	return nil
}

func (v *Vino) StoreComment(ctx context.Context, db *sqlx.DB, comment string) error {
	query := `INSERT INTO comments (content, wine) VALUES (?, ?)`

	_, err := db.ExecContext(ctx, query, comment, v.ID)
	if err != nil {
		return err
	}

	return nil
}

func ListWines(ctx context.Context, db *sqlx.DB) ([]Vino, error) {
	query := `SELECT rowid, name, rating FROM wines ORDER BY rating DESC;`

	var wines []Vino

	err := db.SelectContext(ctx, &wines, query)
	if err != nil {
		return nil, err
	}

	return wines, nil
}
